
import           LCLib

import           Control.Arrow                          ((***))
import           Control.Monad.Reader
import           Control.Monad.Trans                    (liftIO)
import           Data.List                              (group, intercalate,
                                                         isInfixOf, sort)
import qualified Data.Map                               as M
import           Text.Parsec                            hiding (getInput, many,
                                                         space, spaces, (<|>))
import           Text.Parsec.String                     (parseFromFile)
import           Unbound.Generics.LocallyNameless

import           System.Console.Shell
import           System.Console.Shell.Backend.Haskeline
import           System.Console.Shell.ShellMonad

type LCState = Defns

type LC = Sh LCState

lc :: ShellDescription LCState
lc = initialShellDescription
  { shellCommands =
    [ exitCommand "quit"
    , helpCommand "help"
    , cmd "load" loadCmd "Load a definitions file."
    , cmd "save" saveCmd "Save current definitions to a file."
    , cmd "show" showCmd "Show current definitions."
    ]
  , commandStyle = OnlyCommands
  , evaluateFunc = exec
  , greetingText = Just $ unlines
      [ "Welcome to the λ-calculus shell!"
      , ""
      , "Type 'help' to see a list of available commands,"
      , "or enter a lambda-calculus term to reduce."
      , "By default, terms are reduced for at most 100 steps;"
      , "to increase the limit, type n: <term>, where n is the"
      , "number of steps and <term> is the term to be reduced."
      , ""
      ]
  }

saveCmd :: FilePath -> LC ()
saveCmd f = do
  defns <- getShellSt
  liftIO $ writeFile f (showDefns defns)
  shellPutStrLn "Saved."

loadCmd :: FilePath -> LC ()
loadCmd f = do
  res <- liftIO $ parseFromFile parseDefns f
  case res of
    Left parseErr        -> shellPutStrLn (show parseErr)
    Right defns -> do
      let names = map fst defns
          dupes = map head . filter ((>1) . length) . group . sort $ names
      case dupes of
        (_:_) -> shellPutStrLn $ "Duplicate definitions: " ++ intercalate " " dupes
        _     -> do
          let defnsMap = M.fromList defns
          case detectCycleDefns defnsMap of
            Left names -> cycleError names
            Right _    -> do
              shellPutStrLn "Loaded."
              putShellSt defnsMap

cycleError names = shellPutStrLn $ "Illegal cycle in definitions: " ++ intercalate " " names

showDefns defns = unlines . map showDefn . M.assocs $ defns
  where
    showDefn (name,t) = name ++ " = " ++ ppTermLC t

showCmd :: LC ()
showCmd = do
  defns <- getShellSt
  shellPutStr . showDefns $ defns

exec :: String -> LC ()
exec str
  | "=" `isInfixOf` str = execDefn str
  | otherwise           = eval str

execDefn :: String -> LC ()
execDefn str =
  case parse parseDefn "" str of
    Left pe         -> shellPutStrLn (show pe)
    Right (name, t) -> do
      defns <- getShellSt
      let defns' = M.insert name t defns
      case detectCycleDefns defns' of
        Left names -> cycleError names
        Right _    -> putShellSt defns'

eval :: String -> LC ()
eval str = do
  defns <- getShellSt
  let (nsteps, str')
        | ":" `isInfixOf` str = (read *** tail) (break (==':') str)
        | otherwise = (100, str)

  case parse parseLCLine "" str' of
    Left pe -> shellPutStrLn (show pe)
    Right t -> do
      let out = unlines . map ppTermLC . runLFreshM . flip runReaderT defns . steps nsteps
              $ t
      shellPutStrLn ("\n" ++ out)

main = do
  runShell lc haskelineBackend (M.empty)
