{-# LANGUAGE DeriveGeneric         #-}
{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE FlexibleInstances     #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE UndecidableInstances  #-}

module LCLib where

import           Control.Lens              (toListOf)

import           Control.Applicative
import           Control.Monad.Reader
import           Control.Monad.Trans.Maybe
import           Data.List                 (intercalate)
import qualified Data.Map                  as M
import           Test.QuickCheck           hiding (output)
import           Text.Parsec               hiding (getInput, many, space,
                                            spaces, (<|>))
import           Text.Parsec.String        (Parser)

import           GHC.Generics
import           Unbound.Generics.LocallyNameless   hiding (join, name)

------------------------------------------------------------
-- LC types
------------------------------------------------------------

data Term = Var (Name Term)
          | Lam (Bind (Name Term) Term)
          | App Term Term
  deriving (Show, Generic)

instance Alpha Term
instance Subst Term Term where
  isvar (Var v) = Just (SubstName v)
  isvar _       = Nothing

var :: String -> Term
var x = Var (s2n x)

f = var "f"
x = var "x"
y = var "y"
z = var "z"

lam :: String -> Term -> Term
lam x b = Lam $ bind (s2n x) b

app :: Term -> Term -> Term
app = App

type Defn = (String, Term)
type Defns = M.Map String Term
data Prog = Prog Defns Term
  deriving Show

addDefn :: Defn -> Prog -> Parser Prog
addDefn (x,def) (Prog m t)
  | x `M.member` m
    = parserFail $ "Duplicate definition of " ++ x
  | otherwise = return $ Prog (M.insert x def m) t

------------------------------------------------------------
-- Parsing
------------------------------------------------------------

space   = char ' ' <|> char '\t'
spaceNL = space <|> char '\n'
spaces  = many space
spacesNL = many spaceNL
spaces1 = many1 space

parseId :: Parser String
parseId = spaces *> many1 alphaNum <* spaces

parseVar, parseLC, parseLCLam, parseLCAtom :: Parser Term

parseVar = var <$> parseId

parseLC = spaces *> (parseLCLam <|> chainl1 parseLCAtom (spaces *> pure app)) <* spaces

parseLCLam = lam <$> ((char '\\' <|> char 'λ') *> spaces *> parseId)
                 <*> (spaces *> char '.' *> spaces *> parseLC)

parseLCAtom = parseVar <|> parens parseLC

parens = between (char '(') (char ')')

parseAlg :: Parser Term
parseAlg = spaces *>
           ( var <$> (string "Var" *> spaces *> parens (parseId))
         <|> lam <$> (string "Lam" *> spaces *> char '(' *> parseId) <*> (char ',' *> spaces *> parseAlg <* spaces <* char ')')
         <|> app <$> (string "App" *> spaces *> char '(' *> parseAlg) <*> (spaces *> char ',' *> spaces *> parseAlg <* spaces <* char ')')
           )

parseLCAny = try parseLC <|> parseAlg
parseLCLine = parseLCAny <* spaces <* ((pure () <* eol) <|> eof <|> (pure () <* char ';'))
eol = (char '\n' >> return ()) <|> (string "\r\n" >> return ())

parseDefn :: Parser (String, Term)
parseDefn = (,) <$> (spaces *> parseId) <*> (spaces *> char '=' *> spaces *> parseLCLine)

parseDefns :: Parser [Defn]
parseDefns = many (spacesNL *> parseDefn)

parseProg :: Parser Prog
parseProg = Prog M.empty <$> try parseLCLine
        <|> join (addDefn <$> parseDefn <*> parseProg)

detectCycle :: Prog -> Either [String] ()
detectCycle (Prog defns _) = detectCycleDefns defns

detectCycleDefns :: Defns -> Either [String] ()
detectCycleDefns defns = mapM_ (detectCycleFrom []) (M.keys defns)
  where
    detectCycleFrom path cur
      | cur `elem` path = Left path
      | otherwise       = maybe (return ()) (mapM_ (\free -> detectCycleFrom (cur:path) free) . map anyName2String . toListOf fvAny) $ M.lookup cur defns
    anyName2String (AnyName n) = name2String n

{-

| x `elem` (map anyName2String $ fvAny def)
  = parserFail $ "Illegal recursive definition of " ++ x

-}

------------------------------------------------------------
-- Pretty-printing
------------------------------------------------------------

ppTermLC :: Term -> String
ppTermLC = runLFreshM . ppTermLC' 0
  where
    ppTermLC' :: Int -> Term -> LFreshM String
    ppTermLC' _ (Var x) = return (show x)
    ppTermLC' p (Lam l) = do
      lunbind l $ \(x,t) -> do
        t' <- ppTermLC' 0 t
        return $ showParens (p > 0) ("λ" ++ show x ++ ". " ++ t')
    ppTermLC' p (App t1 t2) = do
      t1' <- ppTermLC' 1 t1
      t2' <- ppTermLC' 2 t2
      return $ showParens (p > 1) (t1' ++ " " ++ t2')

showParens True x = "(" ++ x ++ ")"
showParens _    x = x

ppTermAlg :: Term -> String
ppTermAlg = runLFreshM . ppTermAlg'
  where
    ppTermAlg' :: Term -> LFreshM String
    ppTermAlg' (Var x) = return $ cons "Var" [show x]
    ppTermAlg' (Lam l) = do
      lunbind l $ \(x,t) -> do
        t' <- ppTermAlg' t
        return $ cons "Lam" [show x, t']
    ppTermAlg' (App t1 t2) = cons "App" <$> (mapM ppTermAlg' [t1,t2])

cons k args = k ++ "(" ++ intercalate ", " args ++ ")"

------------------------------------------------------------
-- Evaluation
------------------------------------------------------------

step :: Term -> MaybeT (ReaderT Defns LFreshM) Term
step (Var _)         = mzero
step (Lam _)         = mzero
step (App (Var x) t) = do
  mx' <- asks (M.lookup (name2String x))
  case mx' of
    Nothing -> App <$> pure (Var x) <*> step t
    Just x' -> return (App x' t)
step (App (Lam l) t) = do
  lunbind l $ \(x,b) -> return (subst x t b)
step (App t1 t2)     = App <$> step t1 <*> pure t2
                   <|> App <$> pure t1 <*> step t2

steps :: Int -> Term -> ReaderT Defns LFreshM [Term]
steps 0 t = return [t]
steps n t = do
  mt' <- runMaybeT (step t)
  case mt' of
    Nothing -> return [t]
    Just t' -> (t:) <$> steps (n-1) t'

------------------------------------------------------------
-- QC tests
------------------------------------------------------------

instance Arbitrary Term where
  arbitrary = sized term
    where
      v      = (:[]) <$> elements "vwxyz"
      term 0 = var <$> v
      term n = oneof [ lam <$> v <*> term (n-1)
                     , app <$> term (n `div` 2) <*> term (n `div` 2)
                     ]

prop_LC_pp_parse t =
  case parse parseLC "" (ppTermLC t) of
    Left _ -> False
    Right t' -> aeq t t'

prop_Alg_pp_parse t =
  case parse parseAlg "" (ppTermAlg t) of
    Left _ -> False
    Right t' -> aeq t t'

